@@ -171,6 +171,13 @@ var SETTINGS_TYPES = {
             "default",
             "description"
         ]
+    },
+    "list" : {
+        "required-fields": [
+            "type",
+            "default",
+            "columns"
+        ]
     }
 };
 
@@ -258,7 +265,7 @@ XletSettingsBase.prototype = {
         this.bindObject = bindObject;
         this.uuid = uuid;
         if (this._get_is_multi_instance_xlet(this.uuid)) this.instanceId = instanceId;
-        else this.instanceId = this.uuid; 
+        else this.instanceId = this.uuid;
         this.bindings = {};
 
         if (!this._ensureSettingsFiles()) return;
@@ -534,14 +541,26 @@ XletSettingsBase.prototype = {
         }
     },
 
-    _settingChangedRemotely: function(key, value) {
-        let changed = false;
-        if (key in this.settingsData) {
-            if (this.settingsData[key].value === undefined) return;
-            if (this.settingsData[key].value == value) return;
+    _checkSettings: function() {
+        let oldSettings = this.settingsData;
+        try {
+            this.settingsData = this._loadFromFile();
+        } catch(e) {
+            // looks like we're getting a premature signal from the file monitor
+            // we should get another when the file is finished writing
+            return;
+        }
 
-            let oldValue = this.settingsData[key].value;
-            this.settingsData[key].value = value;
+        let changed = false;
+        for (let key in this.settingsData) {
+            if (!this.settingsData[key]
+                || this.settingsData[key].value === undefined
+                || !oldSettings[key]
+                || oldSettings[key].value === undefined) continue;
+
+            let oldValue = oldSettings[key].value;
+            let value = this.settingsData[key].value;
+            if (value == oldValue) continue;
 
             changed = true;
             if (key in this.bindings) {
@@ -581,21 +600,34 @@ XletSettingsBase.prototype = {
         let configDir = Gio.file_new_for_path(configPath);
         if (!configDir.query_exists(null)) configDir.make_directory_with_parents(null);
         this.file = configDir.get_child(this.instanceId + ".json");
+        this.monitor = this.file.monitor_file(Gio.FileMonitorFlags.NONE, null);
         let xletDir = this.ext_type.maps.dirs[this.uuid];
         let templateFile = xletDir.get_child("settings-schema.json");
-        let needsSave = false;
 
         // If the settings have already been installed previously we need to check if the schema
         // has changed and if so, do an upgrade
         if (this.file.query_exists(null)) {
-            this.settingsData = this._loadFromFile();
+            try {
+                this.settingsData = this._loadFromFile();
+            } catch(e) {
+                // Some users with a little bit of know-how may be able to fix the file if it's not too corrupted. Is it worth it to give an option to skip this step?
+                global.logError(e);
+                global.logError("Failed to parse file "+this.file.get_path()+". Attempting to rebuild the settings file for "+this.uuid+".");
+                Main.notify("Unfortunately the settings file for "+this.uuid+" seems to be corrupted. Your settings have been reset to default.");
+            }
+        }
+
+        // if this.settingsData is populated, all we need to do is check for a newer version of the schema and upgrade if we find one
+        // if not, it means either that an installation has not yet occurred, or something when wrong
+        // either way, we need to install
+        if (this.settingsData) {
             if (templateFile.query_exists(null)) {
                 let templateData = Cinnamon.get_file_contents_utf8_sync(templateFile.get_path());
                 let checksum = global.get_md5_for_string(templateData);
 
                 try {
                     if (checksum != this.settingsData.__md5__) this._doUpgrade(templateData, checksum);
-                    needsSave = true;
+                    this._saveToFile();
                 } catch(e) {
                     if (e) global.logError(e);
                     global.logWarning("upgrade failed for " + this.uuid + ": falling back to previous settings");
@@ -622,10 +654,10 @@ XletSettingsBase.prototype = {
                 return false;
             }
 
-            needsSave = true;
+            this._saveToFile();
         }
 
-        if (needsSave) this._saveToFile();
+        if (!this.monitorId) this.monitorId = this.monitor.connect("changed", Lang.bind(this, this._checkSettings));
 
         return true;
     },
@@ -652,8 +684,6 @@ XletSettingsBase.prototype = {
         for (let key in newSettings) {
             let props = newSettings[key];
 
-            // if (!has_required_fields(props, key)) throw null;
-
             if (!("type" in props) || !(props.type in SETTINGS_TYPES)) continue;
             let type = SETTINGS_TYPES[props.type];
 
@@ -699,22 +729,24 @@ XletSettingsBase.prototype = {
     _loadFromFile: function() {
         let rawData = Cinnamon.get_file_contents_utf8_sync(this.file.get_path());
         let json = JSON.parse(rawData);
+
         return json;
     },
 
     _saveToFile: function() {
+        if (this.monitorId) this.monitor.disconnect(this.monitorId);
         let rawData = JSON.stringify(this.settingsData, null, 4);
         let raw = this.file.replace(null, false, Gio.FileCreateFlags.NONE, null);
         let out_file = Gio.BufferedOutputStream.new_sized(raw, 4096);
         Cinnamon.write_string_to_stream(out_file, rawData);
         out_file.close(null);
+        this.monitorId = this.monitor.connect("changed", Lang.bind(this, this._checkSettings));
     },
 
-    // called by cinnamonDBus.js to when the setting is changed remotely.
+    // called by cinnamonDBus.js to when the setting is changed remotely. This is to expedite the
+    // update due to settings changes, as the file monitor has a significant delay.
     remoteUpdate: function(key, payload) {
-        let val = JSON.parse(payload);
-
-        this._settingChangedRemotely(key, val);
+        this._checkSettings();
     },
 
     /**
@@ -728,6 +760,7 @@ XletSettingsBase.prototype = {
         for (let key in this.bindings) {
             this.unbindAll(key);
         }
+        if (this.monitorId) this.monitor.disconnect(this.monitorId);
         this.disconnectAll();
     }
 }
